from linalg import Matrix
from linalg.create import zeroes
from linalg.decompose import lu

__all__ = ["linsolve"]


def linsolve(mat: Matrix, b: Matrix) -> Matrix:
    """solve a system of linear equations using LU decomposition
    
    :param mat: system of linear equations
    :type mat: Matrix
    :param b: column vector to solve for
    :type b: Matrix
    :return: the solution
    :rtype: Matrix
    """
    L, U, P, S = lu(mat)
    x, y = zeroes(*b.shape), zeroes(*b.shape)
    b = P @ b

    n = y.shape[0]

    for i in range(y.shape[1]):
        for j in range(n):  # solve Ly = Pb for y (forward substitution)
            y[j][i] = b[j][i]
            for k in range(j):
                y[j][i] -= L[j][k] * y[k][i]

        for j in range(n - 1, -1, -1):  # solve Ux = y for x (backward substitution)
            x[j][i] = y[j][i]
            for k in range(j + 1, n):
                x[j][i] -= U[j][k] * x[k][i]
            x[j][i] /= U[j][j]

    return x



from operator import add, neg
import linalg


__all__ = ["Matrix"]


class MatrixError(Exception):
    pass


class Matrix:
    """Implements Matrices.
    """

    def __init__(self, mat, valid=False):
        if not valid:
            try:
                self._is_valid(mat)
            except Exception as e:
                raise e

        self.shape = (len(mat), len(mat[0]))
        self.matrix = mat

    def _is_valid(self, mat):
        """checks if given matrix is valid

        :param mat: Matrix to test
        :type mat: Matrix
        :raises MatrixError: raises MatrixError
        """
        matlen = len(mat[0])
        for row in mat:
            if len(row) != matlen:
                raise MatrixError("Malformed matrix")
            for i in row:
                assert isinstance(i, (int, float, complex))

    def _is_square(self) -> bool:
        return self.shape[0] == self.shape[1]

    def inverse(self) -> "Matrix":
        """
        returns the inverse matrix of mat
        implemented as an alias of linalg.unary.inverse

        :param mat: the matrix to invert
        :type mat: Matrix
        :return: the inverse matrix of mat
        :rtype: Matrix
        """
        return linalg.unary.inverse(self)

    def transpose(self) -> "Matrix":
        """
        computes the transpose of self
        implemented as an alias of linalg.unary.transpose    

        :return: transposed matrix
        :rtype: Matrix
        """
        return linalg.unary.transpose(self)

    inv = inverse  # Alias for inverse
    T = transpose  # Alias for transpose

    def det(self) -> float:
        """
        computes the determinant for self.
        implemented as an alias of linalg.unary.det

        :return: the determinant for mat
        :rtype: float
        """

        return linalg.unary.det(self)

    def __str__(self):
        r = ""
        for i in range(self.shape[0]):
            r += "|"
            for j in range(self.shape[1]):
                r += "{:^6.5}".format(str(self[i][j]))
            r += "|"
            r += "\n"

        return r

    def __add__(self, x):
        return Matrix(
            list([list(map(add, self[i], x[i])) for i in range(self.shape[0])]),
            valid=True,
        )

    def __neg__(self):
        return Matrix(
            list([list(map(neg, self[i])) for i in range(self.shape[0])]), valid=True
        )

    def __sub__(self, x):
        return self.__add__(-x)

    def __matmul__(self, x: "Matrix"):
        result = Matrix(
            [
                [sum(a * b for a, b in zip(self_row, x_col)) for x_col in zip(*x)]
                for self_row in self
            ]
        )

        return result

    def __mul__(self, x: float):
        return Matrix([[a * x for a in row] for row in self])

    def __rmul__(self, x: float):
        return self.__mul__(x)

    def __getitem__(self, i):
        return self.matrix[i]

    def __setitem__(self, key, item):
        self.matrix[key] = item



"""Main functions and classes"""


from . import create
from . import decompose
from . import solve
from . import unary
from . import data

from .types import Matrix
from .create import *
from .decompose import *
from .solve import *
from .unary import *
from .data import *

__all__ = ["Matrix"]
__all__.extend(create.__all__)
__all__.extend(decompose.__all__)
__all__.extend(solve.__all__)
__all__.extend(unary.__all__)
__all__.extend(data.__all__)


from linalg import Matrix
import random

__all__ = ["zeroes", "identity", "random_matrix"]


def zeroes(i: int, j: int) -> Matrix:
    """creates an i by j zero matrix
    
    :param i: number of columns
    :type i: int
    :param j: number of rows
    :type j: int
    :return: an i by j matrix filled with zeroes
    :rtype: Matrix
    """
    return Matrix([[0] * j for _ in range(i)], valid=True)


def identity(n: int) -> Matrix:
    """generates an n by n identity matrix
    
    :param n: number of rows/columns
    :type n: int
    :return: n by b indentity matrix
    :rtype: Matrix
    """
    id = [[int(i == j) for j in range(n)] for i in range(n)]
    return Matrix(id)


def random_matrix(dim: tuple, rng: tuple, use_float: bool = False) -> Matrix:
    """generates a random matrix

    :param dim: dimensions of matrix
    :type dim: tuple
    :param rng: range of randomized elements
    :type rng: tuple
    :param float: whether to use floats for elements (default False)
    :type float: bool
    :return: randomized matrix of specified size and range
    :rtype: Matrix
    """
    ls = [[random.uniform(*rng) for _ in range(dim[1])] for _ in range(dim[0])]

    if not use_float:
        ls = [[round(x) for x in row] for row in ls]

    return Matrix(ls)



from linalg import Matrix
from linalg.create import zeroes
import linalg

__all__ = ["lu"]


def lu(mat: Matrix) -> (Matrix, Matrix, Matrix, int):
    """implements LUP decomposition 

    :return: returns a tuple with L, U, and P
    :rtype: Matrix, Matrix, Matrix, int
    """
    assert mat._is_square()

    n = mat.shape[0]

    L, U = zeroes(n, n), zeroes(n, n)
    P, S = linalg.unary.pivotize(mat)
    A2 = P @ mat

    for j in range(n):
        L[j][j] = 1
        for i in range(j + 1):
            s1 = sum(U[k][j] * L[i][k] for k in range(i))
            U[i][j] = A2[i][j] - s1
        for i in range(j, n):
            s2 = sum(U[k][j] * L[i][k] for k in range(j))
            L[i][j] = (A2[i][j] - s2) / U[j][j]

    return L, U, P, S




from linalg import Matrix

__all__ = ["as_matrix", "as_list"]


def as_matrix(ls: list) -> Matrix:
    """
    Parses the input 2D list into a Matrix.
    Convenience top-level function for linalg.matrix.Matrix()

    :param ls: list to convert into matrix
    :type ls: list
    :return: Matrix represtation of ls
    :rtype: Matrix
    """
    return Matrix(ls)


def as_list(mat: Matrix) -> list:
    """
    Returns the matrix as a vanilla 2D list.
    
    :param mat: matrix to convert
    :type mat: Matrix
    :return: matrix as a 2d list
    :rtype: list
    """
    return mat.matrix



from setuptools import setup

NAME = "linalg"
DESCRIPTION = "A simple linear algebra package written in vanilla python3"
URL = "https://github.com/rocketll/linalg"
EMAIL = "luc4sl33@gmail.com"
AUTHOR = "Lucas Lee"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = "1.0.3"


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name=NAME,
    packages=["linalg"],
    version=VERSION,
    description=DESCRIPTION,
    long_description=readme(),
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    include_package_data=True,
    keywords=["linear algebra", "math"],
    license="MIT",
)




from linalg import Matrix
from linalg.create import identity
from linalg.solve import linsolve
from linalg.decompose import lu

__all__ = ["pivotize", "inverse", "det", "transpose", "T", "inv"]


def pivotize(mat: Matrix) -> (Matrix, int):
    """creates the pivoting matrix for mat

    :param mat: the matrix to perform the operation on
    :type mat: Matrix
    :return: the pivoting matrix for self and the number of permutations
    :rtype: Matrix, int
    """
    assert mat._is_square()
    n = mat.shape[0]
    S = 0
    a = identity(n)
    for j in range(n):
        row = max(range(j, n), key=lambda i: abs(mat[i][j]))
        if j != row:
            S += 1
            a[j], a[row] = a[row], a[j]

    return a, S


def inverse(mat: Matrix) -> Matrix:
    """returns the inverse matrix of mat
    
    :param mat: the matrix to invert
    :type mat: Matrix
    :return: the inverse matrix of mat
    :rtype: Matrix
    """
    assert mat._is_square()
    return linsolve(mat, identity(mat.shape[0]))


def det(mat: Matrix) -> float:
    """computes the determinant for a given matrix
    
    :param mat: matrix to compute the determinant for
    :type mat: Matrix
    :return: the determinant for mat
    :rtype: float
    """
    assert mat._is_square()

    L, U, P, S = lu(mat)
    n = mat.shape[0]
    l_pd, u_pd = 1, 1
    for i in range(n):
        l_pd *= L[i][i]
        u_pd *= U[i][i]
    return (-1) ** S * l_pd * u_pd


def transpose(mat: Matrix) -> Matrix:
    """
    computes the transpose of a given matrix
    
    :param mat: matrix to perform the operation on
    :type mat: Matrix
    :return: transposed matrix
    :rtype: Matrix
    """
    x, y = mat.shape
    return Matrix([[mat[a][b] for a in range(y)] for b in range(x)])


T = transpose  # Function alias
inv = inverse  # Function alias
